<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 11/16/2003 *
 * ===================                          Last change: 12/13/2004 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : mysql-connect.php                                *
 * -------------------------------------------------------------------- *
 * Short description : Connects to your database                        *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Verbindet zu Ihrer Datenbank                     *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2011 by Mailer Developer Team                   *
 * For more information visit: http://www.mxchange.org                  *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// Load more function libraries or includes
foreach (array('request-functions', 'session-functions', 'code-functions', 'language-functions', 'sql-functions', 'expression-functions', 'filter-functions','revision-functions', 'filters', 'mysql-manager', 'extensions-functions', 'handler') as $lib) {
	// Load special functions
	loadIncludeOnce('inc/' . $lib . '.php');
} // END - foreach

// Set error handler
set_error_handler('__errorHandler');

// Disable block-mode by default
enableBlockMode(false);

// Init error handler
initErrorHandler();

// Init request
initRequest();

// Init userid
initMemberId();

// Set important header_sent
if (!isset($GLOBALS['header_sent'])) $GLOBALS['header_sent'] = '0';

// Init fatal messages
initFatalMessages();

// Init repository data sub-system
initRepositoryData();

// Enable HTML templates by default
enableTemplateHtml();

// Are we in installation phase?
if ((!isInstalling()) && (!isInstallationPhase())) {
	// Load configuration file(s) here
	loadIncludeOnce('inc/load_config.php');

	// Load database layer here
	loadIncludeOnce('inc/db/lib.php');

	// Init message system
	initMessages();

	// CSS array
	initExtensionCssFiles();

	if ((!empty($GLOBALS['mysql']['host'])) && (!empty($GLOBALS['mysql']['login'])) && (!empty($GLOBALS['mysql']['dbase']))) {
		// Connect to DB
		SQL_CONNECT($GLOBALS['mysql']['host'], $GLOBALS['mysql']['login'], $GLOBALS['mysql']['password'], __FILE__, __LINE__);

		// Is the link valid?
		if (SQL_IS_LINK_UP()) {
			// Enable exit on error
			enableExitOnError();

			// Is it a valid resource?
			if (SQL_SELECT_DB($GLOBALS['mysql']['dbase'], __FILE__, __LINE__) === true) {
				// This is required for extension 'optimize' to work
				setConfigEntry('__DB_NAME', $GLOBALS['mysql']['dbase']);

				// Remove MySQL array from namespace
				unset($GLOBALS['mysql']);

				// Load cache
				loadIncludeOnce('inc/load_cache.php');
			} else {
				// Wrong database?
				addFatalMessage(__FILE__, __LINE__, '{--WRONG_DB_SELECTED--}');
			}
		} else {
			// No link to database!
			addFatalMessage(__FILE__, __LINE__, '{--NO_DB_LINK--}');
		}
	} else {
		// Maybe you forgot to enter your database login?
		addFatalMessage(__FILE__, __LINE__, '{--MYSQL_DATA_MISSING--}');
	}
} else {
	// Default output is 'direct' for HTML output
	setConfigEntry('OUTPUT_MODE', 'direct');

	// This hack prevents a backtrace in CSS output
	if (isCssOutputMode()) {
		// Problem with config so set output mode
		setConfigEntry('OUTPUT_MODE', 'render');
	} // END - if

	// CFG: DATABASE-TYPE
	setConfigEntry('_DB_TYPE', 'mysql3');

	// Load database layer here
	loadIncludeOnce('inc/db/lib.php');

	// Init message system
	initMessages();

	// Include more
	foreach (array('databases', 'session', 'versions', 'install-functions', 'load_config', 'load_cache') as $inc) {
		// Load the include
		loadIncludeOnce('inc/' . $inc . '.php');
	} // END - foreach

	// Check wether we are in installation routine
	if ((!isInstalling()) && (!isCssOutputMode()) && (!isRawOutputMode())) {
		// Redirect to the URL
		redirectToUrl('install.php');
	} // END - if
}

// Handle fatal errors
runFilterChain('handle_fatal_errors');

// [EOF]
?>
